// Copyright 2024. The Tari Project
//
// Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
// following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
// disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
// following disclaimer in the documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
// products derived from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
// INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
// USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

use std::{path::PathBuf, time::{SystemTime, UNIX_EPOCH}};
use tokio::time::sleep;

use anyhow::{anyhow, Result};
use clap::Parser;
use log::*;
use reqwest::Client;
use serde::{Deserialize, Serialize};
use tonic::transport::{Certificate, ClientTlsConfig, Endpoint};
use zmq::{Context, Socket};

use minotari_app_grpc::{
    authentication::ClientAuthenticationInterceptor,
    conversions::transaction_output::grpc_output_with_payref,
    tari_rpc::{
        base_node_client::BaseNodeClient, pow_algo::PowAlgos, Block, NewBlockTemplateRequest, PowAlgo,
        TransactionInput, TransactionOutput, TransactionKernel,
    },
};
use minotari_app_utilities::parse_miner_input::BaseNodeGrpcClient;
use std::str::FromStr;
use tari_common::configuration::Network;
use tari_common::MAX_GRPC_MESSAGE_SIZE;
use tari_common_types::{grpc_authentication::GrpcAuthentication, tari_address::TariAddress};
use tari_core::{
    consensus::ConsensusManager,
    transactions::{
        generate_coinbase,
        tari_amount::MicroMinotari,
        transaction_components::{
            encrypted_data::{PaymentId, TxType},
            CoinBaseExtra, RangeProofType,
        },
        transaction_key_manager::{create_memory_db_key_manager, MemoryDbKeyManager},
    },
};
use tari_utilities::hex::Hex;
use tari_hashing::DomainSeparatedBorshHasher;
use blake2::Blake2b;
use digest::consts::U32;
use tari_core::blocks::BlocksHashDomain;

const LOG_TARGET: &str = "gbt::main";

// 区块头结构
#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BlockHeader {
    pub hash: String,
    pub version: u32,
    pub height: u64,
    pub prev_hash: String,
    pub timestamp: u64,
    pub output_mr: String,
    pub block_output_mr: String,
    pub kernel_mr: String,
    pub input_mr: String,
    pub total_kernel_offset: String,
    pub nonce: u64,
    pub pow: ProofOfWork,
    pub kernel_mmr_size: u64,
    pub output_mmr_size: u64,
    pub total_script_offset: String,
    pub validator_node_mr: String,
    pub validator_node_size: u64,
    pub output_smt_size: u64,  // 新增：output_smt_size字段
}

// 工作量证明结构
#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct ProofOfWork {
    pub pow_algo: u64,
    pub pow_data: String,
}

// 区块体结构
#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct BlockBody {
    pub inputs: Vec<TransactionInput>,
    pub outputs: Vec<TransactionOutput>,
    pub kernels: Vec<TransactionKernel>,
}

// 挖矿任务结构
#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct MiningTask {
    pub block_header: BlockHeader,
    pub block_body: BlockBody,
    pub output_smt_size: u64,
    pub coinbase_hash: String,
    pub target: u64,  // 新增：目标难度
    pub created_at: u64,  // 新增：创建时间戳（Unix时间戳）
}

// 挖矿消息结构
#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct MiningMsg {
    pub height: u64,
    pub mining_hash: String,
    pub target: u64,
    pub block_header: BlockHeader,
    pub block_body: BlockBody,
}

// 提交请求结构
#[derive(Debug, Serialize, Deserialize)]
pub struct SubmitRequest {
    pub job_id: String,
    pub nonce: u64,
    pub solution: String,
}

// 提交结果结构
#[derive(Debug, Serialize, Deserialize)]
pub struct SubmitResult {
    pub job_id: String,
    pub result: u8, // 1表示成功，0表示失败
}

// === 新增：output_smt_size 计算方法 ===
fn calculate_output_smt_size(
    prev_output_smt_size: u64,
    outputs: &[String],  // JSON字符串格式
    inputs: &[String],   // JSON字符串格式
) -> u64 {
    let mut size = prev_output_smt_size;
    let mut new_leaves = 0u64;
    let mut stale_leaves = 0u64;
    
    // 计算新叶子数量（排除销毁输出）
    for (i, output_json) in outputs.iter().enumerate() {
        match serde_json::from_str::<serde_json::Value>(output_json) {
            Ok(output) => {
                // 尝试多种可能的JSON结构
                let is_burned = if let Some(features) = output.get("features") {
                    if let Some(output_type) = features.get("output_type") {
                        output_type.as_u64() == Some(2) // Burn = 2
                    } else {
                        false
                    }
                } else if let Some(output_type) = output.get("output_type") {
                    output_type.as_u64() == Some(2) // Burn = 2
                } else {
                    false
                };
                
                if !is_burned {
                    new_leaves += 1;
                    debug!(target: LOG_TARGET, "Output {}: not burned, adding leaf", i);
                } else {
                    debug!(target: LOG_TARGET, "Output {}: burned, skipping", i);
                }
            },
            Err(e) => {
                warn!(target: LOG_TARGET, "Failed to parse output {} JSON: {}, assuming not burned", i, e);
                new_leaves += 1; // 保守策略：假设不是销毁输出
            }
        }
    }
    
    // 计算移除叶子数量（排除销毁输入）
    for (i, input_json) in inputs.iter().enumerate() {
        match serde_json::from_str::<serde_json::Value>(input_json) {
            Ok(input) => {
                // 尝试多种可能的JSON结构
                let is_burned = if let Some(features) = input.get("features") {
                    if let Some(output_type) = features.get("output_type") {
                        output_type.as_u64() == Some(2) // Burn = 2
                    } else {
                        false
                    }
                } else if let Some(output_type) = input.get("output_type") {
                    output_type.as_u64() == Some(2) // Burn = 2
                } else {
                    false
                };
                
                if !is_burned {
                    stale_leaves += 1;
                    debug!(target: LOG_TARGET, "Input {}: not burned, removing leaf", i);
                } else {
                    debug!(target: LOG_TARGET, "Input {}: burned, skipping", i);
                }
            },
            Err(e) => {
                warn!(target: LOG_TARGET, "Failed to parse input {} JSON: {}, assuming not burned", i, e);
                stale_leaves += 1; // 保守策略：假设不是销毁输入
            }
        }
    }
    
    size += new_leaves;
    size = size.saturating_sub(stale_leaves);
    
    info!(target: LOG_TARGET, "output_smt_size calculation: {} (prev) + {} (new_leaves) - {} (stale_leaves) = {} (current)", 
          prev_output_smt_size, new_leaves, stale_leaves, size);
    
    size
}

// === 新增：mining_hash 计算方法 ===
fn calculate_mining_hash(header: &BlockHeader) -> Result<String> {
    let mut hasher = DomainSeparatedBorshHasher::<BlocksHashDomain, Blake2b<U32>>::new_with_label("block_header");
    
    hasher.update_consensus_encode(&header.version);
    hasher.update_consensus_encode(&header.height);
    hasher.update_consensus_encode(&header.prev_hash);
    hasher.update_consensus_encode(&header.timestamp);
    hasher.update_consensus_encode(&header.input_mr);
    hasher.update_consensus_encode(&header.output_mr);
    hasher.update_consensus_encode(&header.output_smt_size);
    hasher.update_consensus_encode(&header.block_output_mr);
    hasher.update_consensus_encode(&header.kernel_mr);
    hasher.update_consensus_encode(&header.kernel_mmr_size);
    hasher.update_consensus_encode(&header.total_kernel_offset);
    hasher.update_consensus_encode(&header.total_script_offset);
    hasher.update_consensus_encode(&header.validator_node_mr);
    hasher.update_consensus_encode(&header.validator_node_size);
    
    let mining_hash = hasher.finalize();
    Ok(hex::encode(mining_hash.as_slice()))
}

// 配置结构
#[derive(Debug, Clone)]
pub struct GbtConfig {
    pub base_node_grpc_address: String,
    pub base_node_http_address: String,
    pub base_node_grpc_authentication: GrpcAuthentication,
    pub base_node_grpc_tls_domain_name: Option<String>,
    pub base_node_grpc_ca_cert_filename: Option<String>,
    pub config_dir: PathBuf,
    pub network: Network,
    pub wallet_payment_address: String,
    pub coinbase_extra: String,
    pub range_proof_type: RangeProofType,
    pub zmq_publisher_port: u16,
    pub zmq_subscriber_port: u16,
}

// GBT客户端
pub struct GbtClient {
    base_node_client: BaseNodeGrpcClient,
    http_client: Client,
    key_manager: MemoryDbKeyManager,
    consensus_manager: ConsensusManager,
    wallet_payment_address: TariAddress,
    config: GbtConfig,

    // ZMQ相关
    #[allow(dead_code)]
    zmq_context: Context,
    publisher_socket: Socket,
}

impl GbtClient {
    pub async fn new(config: GbtConfig) -> Result<Self> {
        // 创建BaseNode客户端
        let base_node_client = Self::connect_base_node(&config).await?;

        // 创建HTTP客户端
        let http_client = Client::new();

        // 创建密钥管理器
        let key_manager = create_memory_db_key_manager().map_err(|e| anyhow!("Key manager error: {}", e))?;

        // 创建共识管理器
        let consensus_manager = ConsensusManager::builder(config.network)
            .build()
            .map_err(|e| anyhow!("Consensus manager error: {}", e))?;

        // 解析钱包地址
        let wallet_payment_address = TariAddress::from_str(&config.wallet_payment_address)
            .map_err(|e| anyhow!("Invalid wallet address: {}", e))?;

        // 创建ZMQ上下文和套接字
        let zmq_context = Context::new();
        let publisher_socket = zmq_context
            .socket(zmq::PUB)
            .map_err(|e| anyhow!("ZMQ publisher error: {}", e))?;

        // 绑定ZMQ套接字
        let publisher_addr = format!("tcp://*:{}", config.zmq_publisher_port);

        publisher_socket
            .bind(&publisher_addr)
            .map_err(|e| anyhow!("ZMQ bind error: {}", e))?;

        Ok(Self {
            base_node_client,
            http_client,
            key_manager,
            consensus_manager,
            wallet_payment_address,
            config,
            zmq_context,
            publisher_socket,
        })
    }

    // 连接BaseNode
    async fn connect_base_node(config: &GbtConfig) -> Result<BaseNodeGrpcClient> {
        info!(target: LOG_TARGET, "Connecting to base node at {}", config.base_node_grpc_address);
        let address = format!("http://{}", config.base_node_grpc_address);
        let mut endpoint = Endpoint::new(address)?;

        // 配置TLS（如果需要）
        if let Some(domain_name) = config.base_node_grpc_tls_domain_name.as_ref() {
            if let Some(cert_filename) = config.base_node_grpc_ca_cert_filename.as_ref() {
                let cert_path = config.config_dir.join(cert_filename);
                let pem = tokio::fs::read(cert_path)
                    .await
                    .map_err(|e| anyhow!("TLS certificate read error: {}", e))?;
                let ca = Certificate::from_pem(pem);

                let tls = ClientTlsConfig::new().ca_certificate(ca).domain_name(domain_name);
                endpoint = endpoint
                    .tls_config(tls)
                    .map_err(|e| anyhow!("TLS config error: {}", e))?;
            }
        }

        let channel = endpoint
            .connect()
            .await
            .map_err(|e| anyhow!("Connection error: {}", e))?;

        let node_conn = BaseNodeClient::with_interceptor(
            channel,
            ClientAuthenticationInterceptor::create(&config.base_node_grpc_authentication)
                .map_err(|e| anyhow!("Authentication error: {}", e))?,
        )
        .max_encoding_message_size(MAX_GRPC_MESSAGE_SIZE)
        .max_decoding_message_size(MAX_GRPC_MESSAGE_SIZE);

        Ok(node_conn)
    }

    // 将gRPC Block转换为自定义结构
    fn convert_block_to_structures(&self, block: &Block) -> Result<(BlockHeader, BlockBody)> {
        let header = block.header.as_ref()
            .ok_or_else(|| anyhow!("No header in block"))?;
        
        let body = block.body.as_ref()
            .ok_or_else(|| anyhow!("No body in block"))?;

        // 构造BlockHeader
        let block_header = BlockHeader {
            hash: header.hash.to_hex(),
            version: header.version,
            height: header.height,
            prev_hash: header.prev_hash.to_hex(),
            timestamp: header.timestamp,
            output_mr: header.output_mr.to_hex(),
            block_output_mr: header.block_output_mr.to_hex(),
            kernel_mr: header.kernel_mr.to_hex(),
            input_mr: header.input_mr.to_hex(),
            total_kernel_offset: header.total_kernel_offset.to_hex(),
            nonce: 0, // 初始为0，等待后续修改
            pow: ProofOfWork {
                pow_algo: header.pow.as_ref().map(|p| p.pow_algo).unwrap_or(0),
                pow_data: header.pow.as_ref().map(|p| p.pow_data.to_hex()).unwrap_or_default(),
            },
            kernel_mmr_size: header.kernel_mmr_size,
            output_mmr_size: header.output_mmr_size,
            total_script_offset: header.total_script_offset.to_hex(),
            validator_node_mr: header.validator_node_mr.to_hex(),
            validator_node_size: header.validator_node_size,
            output_smt_size: header.output_mmr_size, // 使用从gRPC获取的output_mmr_size作为初始值
        };

        // 构造BlockBody - 直接克隆原始结构体
        let block_body = BlockBody {
            inputs: body.inputs.clone(),
            outputs: body.outputs.clone(),
            kernels: body.kernels.clone(),
        };

        Ok((block_header, block_body))
    }

    pub async fn get_block_template_and_coinbase(&mut self) -> Result<MiningTask> {
        info!(target: LOG_TARGET, "Getting new block template");

        // 获取区块模板
        let pow_algo = PowAlgo {
            pow_algo: PowAlgos::Sha3x.into(),
        };

        let request = NewBlockTemplateRequest {
            algo: Some(pow_algo),
            max_weight: 0,
        };

        let template_response = self
            .base_node_client
            .get_new_block_template(request)
            .await?
            .into_inner();

        let mut block_template = template_response
            .new_block_template
            .clone()
            .ok_or_else(|| anyhow!("No block template received"))?;

        let height = block_template
            .header
            .as_ref()
            .ok_or_else(|| anyhow!("No header in block template"))?
            .height;

        // 获取挖矿数据
        let miner_data = template_response
            .miner_data
            .ok_or_else(|| anyhow!("No miner data received"))?;

        let fee = MicroMinotari::from(miner_data.total_fees);
        let reward = MicroMinotari::from(miner_data.reward);
        let target_difficulty = miner_data.target_difficulty;  // 获取目标难度

        info!(target: LOG_TARGET, "Generating coinbase for height {} with target difficulty {}", height, target_difficulty);

        // 生成coinbase
        let (coinbase_output, coinbase_kernel) = generate_coinbase(
            fee,
            reward,
            height,
            &CoinBaseExtra::try_from(self.config.coinbase_extra.as_bytes().to_vec())?,
            &self.key_manager,
            &self.wallet_payment_address,
            true,
            self.consensus_manager.consensus_constants(height),
            self.config.range_proof_type,
            PaymentId::Open {
                user_data: vec![],
                tx_type: TxType::Coinbase,
            },
        )
        .await
        .map_err(|e| anyhow!("Coinbase generation error: {}", e))?;

        // 将coinbase添加到区块模板
        let body = block_template
            .body
            .as_mut()
            .ok_or_else(|| anyhow!("No body in block template"))?;

        let grpc_output = grpc_output_with_payref(coinbase_output.clone(), None)
            .map_err(|e| anyhow!("Output conversion error: {}", e))?;

        body.outputs.push(grpc_output);
        body.kernels.push(coinbase_kernel.into());

        // 获取完整的区块
        let block_result = self.base_node_client.get_new_block(block_template.clone()).await?.into_inner();
        let block = block_result.block.ok_or_else(|| anyhow!("No block in response"))?;

        // 计算coinbase哈希
        let coinbase_hash = coinbase_output.hash().to_hex();

        // 转换为自定义结构
        let (mut block_header, block_body) = self.convert_block_to_structures(&block)?;

        // 先将 outputs 和 inputs 序列化为 JSON 字符串
        let output_jsons: Vec<String> = block_body.outputs.iter()
            .map(|o| serde_json::to_string(o).unwrap_or_default())
            .collect();
        let input_jsons: Vec<String> = block_body.inputs.iter()
            .map(|i| serde_json::to_string(i).unwrap_or_default())
            .collect();
        let current_output_smt_size = calculate_output_smt_size(0, &output_jsons, &input_jsons);

        // 更新block_header中的output_smt_size
        block_header.output_smt_size = current_output_smt_size;

        let mining_task = MiningTask {
            block_header: block_header.clone(),
            block_body,
            output_smt_size: current_output_smt_size,
            coinbase_hash,
            target: target_difficulty,
            created_at: SystemTime::now().duration_since(UNIX_EPOCH).unwrap().as_secs(),
        };

        Ok(mining_task)
    }

    // 发送挖矿消息
    pub fn send_mining_msg(&self, block_header: &BlockHeader, block_body: &BlockBody, target: u64) -> Result<()> {
        let mining_msg = MiningMsg {
            height: block_header.height,
            mining_hash: calculate_mining_hash(block_header)?,
            target,
            block_header: block_header.clone(),
            block_body: block_body.clone(),
        };
        let msg_json = serde_json::to_string(&mining_msg).map_err(|e| anyhow!("Serialization error: {}", e))?;
        self.publisher_socket
            .send_multipart(&["mining_msg".as_bytes(), msg_json.as_bytes()], 0)
            .map_err(|e| anyhow!("ZMQ send error: {}", e))?;
        info!(target: LOG_TARGET, "Sent mining message for height {}, target {}", block_header.height, target);
        Ok(())
    }

    pub async fn run(&mut self) -> Result<()> {
        info!(target: LOG_TARGET, "Starting GBT client");
        let mut last_height: Option<u64> = None;
        let mut last_send_time: u64 = 0;
        loop {
            // 1. 每秒请求一次模板
            let mining_task = match self.get_block_template_and_coinbase().await {
                Ok(task) => task,
                Err(e) => {
                    error!(target: LOG_TARGET, "Failed to get block template: {}", e);
                    sleep(std::time::Duration::from_secs(1)).await;
                    continue;
                }
            };
            let current_height = mining_task.block_header.height;
            let now = SystemTime::now().duration_since(UNIX_EPOCH).unwrap().as_secs();
            let mut should_send = false;
            match last_height {
                None => {
                    should_send = true;
                    info!(target: LOG_TARGET, "First template, sending immediately");
                },
                Some(prev_height) => {
                    if current_height != prev_height {
                        should_send = true;
                        info!(target: LOG_TARGET, "Height changed: {} -> {}, sending immediately", prev_height, current_height);
                    } else if now >= last_send_time + 5 {
                        should_send = true;
                        info!(target: LOG_TARGET, "Same height {}, sending after 5 seconds", current_height);
                    }
                }
            }
            if should_send {
                if let Err(e) = self.send_mining_msg(&mining_task.block_header, &mining_task.block_body, mining_task.target) {
                    error!(target: LOG_TARGET, "Failed to send mining message: {}", e);
                } else {
                    last_height = Some(current_height);
                    last_send_time = now;
                }
            }
            sleep(std::time::Duration::from_secs(1)).await;
        }
    }
}

impl Drop for GbtClient {
    fn drop(&mut self) {
        info!(target: LOG_TARGET, "GBT client shutting down");
        // ZMQ套接字会在Context销毁时自动关闭
    }
}

#[derive(Parser)]
#[command(author, version, about, long_about = None)]
struct Args {
    /// BaseNode gRPC address
    #[arg(short, long, default_value = "127.0.0.1:18102")]
    base_node: String,

    /// BaseNode HTTP address
    #[arg(long, default_value = "127.0.0.1:9000")]
    base_node_http: String,

    /// Network (mainnet, nextnet, testnet)
    #[arg(short, long, default_value = "mainnet")]
    network: String,

    /// Wallet payment address
    #[arg(
        short,
        long,
        default_value = "14H4atSbXqSLFHDvhjx83ASCJDv3iCDu4T6DotCiCVCYq67koEJbgcbmYpeBpRjcZdRYtJ5CDw9gWRNXpe8chfnQSVU"
    )]
    wallet_address: String,

    /// Coinbase extra data
    #[arg(short, long, default_value = "m2pool.com")]
    coinbase_extra: String,

    /// ZMQ publisher port
    #[arg(long, default_value = "31000")]
    zmq_pub_port: u16,

    /// ZMQ subscriber port
    #[arg(long, default_value = "31001")]
    zmq_sub_port: u16,

    /// Enable TLS
    #[arg(long)]
    tls: bool,

    /// TLS domain name
    #[arg(long)]
    tls_domain: Option<String>,

    /// TLS CA certificate file
    #[arg(long)]
    tls_ca_cert: Option<String>,

    /// Config directory
    #[arg(long, default_value = ".")]
    config_dir: String,
}

#[tokio::main]
async fn main() -> Result<()> {
    // 初始化日志
    env_logger::init();

    let args = Args::parse();

    // 解析网络
    let network = match args.network.as_str() {
        "mainnet" => Network::MainNet,
        "nextnet" => Network::NextNet,
        "testnet" => Network::NextNet, // 使用NextNet作为testnet
        _ => return Err(anyhow!("Invalid network: {}", args.network)),
    };

    // 创建配置
    let config = GbtConfig {
        base_node_grpc_address: args.base_node,
        base_node_http_address: args.base_node_http,
        base_node_grpc_authentication: GrpcAuthentication::None,
        base_node_grpc_tls_domain_name: args.tls_domain,
        base_node_grpc_ca_cert_filename: args.tls_ca_cert,
        config_dir: PathBuf::from(args.config_dir),
        network,
        wallet_payment_address: args.wallet_address,
        coinbase_extra: args.coinbase_extra,
        range_proof_type: RangeProofType::BulletProofPlus,
        zmq_publisher_port: args.zmq_pub_port,
        zmq_subscriber_port: args.zmq_sub_port,
    };

    info!(target: LOG_TARGET, "Starting GBT client with network: {:?}", network);

    // 创建GBT客户端
    let mut client = GbtClient::new(config).await?;

    // 运行客户端
    client.run().await?;

    Ok(())
} 